from typing import List

from .utils import Utils


class Split:
    def __init__(self, api_key: str = None):
        """
        Initializes the Split object with default or user-defined thresholds.
        """
        self.utils = Utils(api_key=api_key)

    def _split_large_chunk(self, paragraphs: List[str], embeddings: List[List[float]]) -> (List[str], List[str]):
        """
        Splits 'paragraphs' by finding the least similar boundary using 'embeddings'
        (which are precomputed for these paragraphs only). Returns (left_part, right_part).
        """
        # If there are 0 or 1 paragraphs, no need to split
        if len(paragraphs) < 2:
            return paragraphs, []

        # We'll assume 'embeddings' is already the same length as 'paragraphs'.
        if len(embeddings) < 2:
            # Can't compute consecutive similarities with fewer than 2 embeddings
            return paragraphs, []

        # Find the least similar consecutive boundary
        split_index = self.utils.get_consecutive_least_similar(embeddings)

        left_part = paragraphs[:split_index + 1]
        right_part = paragraphs[split_index + 1:]
        return left_part, right_part

    def split(
        self,
        extracted_text: str,
        min_chunk_size: int = 4000,
        max_chunk_size: int = 7000
    ) -> List[str]:
        """
        Splits the input text into chunks of token-size between [min_chunk_size, max_chunk_size].
        Once a chunk is in that range, we find the "least similar" boundary, store the left side,
        and re-insert the right side for further splitting.
        """
        paragraphs = [p.strip() for p in extracted_text.split("\n\n") if p.strip()]
        if not paragraphs:
            return []

        # Precompute once
        paragraphs_tokens = [self.utils.count_tokens(p) for p in paragraphs]
        paragraphs_embeddings = self.utils.list_embeddings(paragraphs)

        final_chunks = []
        idx = 0
        n = len(paragraphs)

        while idx < n:
            chunk_paragraphs = []
            chunk_embeddings = []
            chunk_tokens = []  # Keep track of tokens in this chunk
            current_tokens = 0

            # 1) Accumulate until we at least exceed min_chunk_size or run out
            while idx < n and current_tokens < min_chunk_size:
                if current_tokens + paragraphs_tokens[idx] <= max_chunk_size:
                    chunk_paragraphs.append(paragraphs[idx])
                    chunk_embeddings.append(paragraphs_embeddings[idx])
                    chunk_tokens.append(paragraphs_tokens[idx])
                    current_tokens += paragraphs_tokens[idx]
                    idx += 1
                else:
                    # This paragraph alone might exceed max_chunk_size -> handle as you see fit
                    break

            # If we haven't hit min_chunk_size but are out of paragraphs, store remainder and quit
            if current_tokens < min_chunk_size and idx >= n:
                if chunk_paragraphs:
                    final_chunks.append(" ".join(chunk_paragraphs))
                break

            # 2) Keep adding while we're under max_chunk_size
            while idx < n:
                if current_tokens + paragraphs_tokens[idx] <= max_chunk_size:
                    chunk_paragraphs.append(paragraphs[idx])
                    chunk_embeddings.append(paragraphs_embeddings[idx])
                    chunk_tokens.append(paragraphs_tokens[idx])
                    current_tokens += paragraphs_tokens[idx]
                    idx += 1
                else:
                    break

            # Now we have between min_chunk_size and max_chunk_size tokens in 'chunk_paragraphs'
            if chunk_paragraphs:
                # 3) Split at the "least similar" boundary
                left_part, right_part = self._split_large_chunk(
                    chunk_paragraphs, chunk_embeddings
                )

                # We'll figure out how many paragraphs ended up in the left part
                used_count = len(left_part)
                leftover_count = len(right_part)

                # Store left side
                final_chunks.append(" ".join(left_part))

                # If there's leftover, reinsert it into the main lists
                if leftover_count > 0:
                    # Slices for leftover
                    leftover_embeddings = chunk_embeddings[used_count:]
                    leftover_tokens = chunk_tokens[used_count:]

                    # Re-insert them at index=idx
                    paragraphs[idx:idx] = right_part
                    paragraphs_embeddings[idx:idx] = leftover_embeddings
                    paragraphs_tokens[idx:idx] = leftover_tokens

                    # Recompute n, in case the paragraphs list has grown
                    n = len(paragraphs)

        return final_chunks


import uuid

import chromadb
import numpy as np

from .utils import Utils


class Vectordb():
    def __init__(self, path, api_key, collection_name):
        self.client = chromadb.PersistentClient(path = path)
        self.utils = Utils(api_key=api_key)
        self.collection = self.client.get_or_create_collection(collection_name)

    def _convert_numpy(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        elif isinstance(obj, list):
            return [self._convert_numpy(v) for v in obj]
        elif isinstance(obj, dict):
            return {k: self._convert_numpy(v) for k, v in obj.items()}
        return obj

    def save_index(self, tree_index):
        try:
            # Generate a unique file_id if not provided
            file_id = str(uuid.uuid4())

            # Prepare parent (summary) entry
            parent_entry = {
                "id": f"summary_{file_id}",
                "embedding": tree_index["embeddings"],
                "metadata": {
                    "type": "summary",
                    "file_name": tree_index["file_name"],
                    "file_id": file_id,
                    "summary": tree_index["summary"],
                    "keywords": "\n".join(tree_index.get("keywords", [])),
                }
            }

            # Prepare child (chunk) entries
            chunk_entries = [
                {
                    "id": f"chunk_{file_id}_{chunk['id']}",
                    "embedding": chunk["embeddings"],
                    "metadata": {
                        "type": "chunk",
                        "file_id": file_id,
                        "summary": chunk["summary"],
                        "keywords": "\n".join(chunk.get("keywords", [])),
                        "original_chunk": chunk.get("original_chunk", "")
                    }
                }
                for chunk in tree_index.get("chunks", [])
            ]

            all_entries = [parent_entry] + chunk_entries

            self.collection.add(
                ids=[entry["id"] for entry in all_entries],
                embeddings=[entry["embedding"] for entry in all_entries],
                metadatas=[entry["metadata"] for entry in all_entries],
            )

            return {
                "message": "Index saved successfully",
                "file_id": file_id,
            }

        except Exception as e:
            print(f"[save_index error] {e}")
            return {
                "message": "Failed to save index",
                "error": str(e),
            }
        
    def get_summary(self, file_id):
        """
        get the summary in the collection.
        """
        try:
            raw_result = self.collection.get(
                where={"$and": [{"type": "summary"}, {"file_id": file_id}]},
                # include=["distances", "metadatas"]
            )
            
            ids = raw_result.get("ids", [])
            metadatas = raw_result.get("metadatas", [])

            result_length = min(len(ids), len(metadatas))

            flat_result = [
                {
                    "id": self._convert_numpy(ids[i]),
                    "metadata": self._convert_numpy(metadatas[i])
                }
                for i in range(result_length)
            ]

            return {
                "summary": flat_result,
            }
        except Exception as e:
            print(f"[list_index error]: {e}")
            return {
                "message": "Failed to list index",
                "error": str(e),
            }

    def list_summaries(self):
        """
        List all the summaries in the collection.
        """
        try:
            raw_result = self.collection.get(
                where={"type": "summary"},
                # include=["distances", "metadatas"]
            )

            # print(raw_result)
            
            ids = raw_result.get("ids", [])
            metadatas = raw_result.get("metadatas", [])

            result_length = min(len(ids), len(metadatas))

            flat_result = [
                {
                    "id": self._convert_numpy(ids[i]),
                    "metadata": self._convert_numpy(metadatas[i])
                }
                for i in range(result_length)
            ]

            return {
                "summaries": flat_result,
            }
        except Exception as e:
            print(f"[list_index error]: {e}")
            return {
                "message": "Failed to list index",
                "error": str(e),
            }
        
    def list_chunks(self, file_id):
        """
        List all the chunks in the collection.
        """

        try:
            raw_result = self.collection.get(
                where={"$and": [{"type": "chunk"}, {"file_id": file_id}]},
            )

            ids = raw_result.get("ids", [])
            metadatas = raw_result.get("metadatas", [])

            result_length = min(len(ids), len(metadatas))

            flat_result = [
                {
                    "id": self._convert_numpy(ids[i]),
                    "metadata": self._convert_numpy(metadatas[i])
                }
                for i in range(result_length)
            ]

            return {
                "chunks": flat_result,
            }
        except Exception as e:
            print(f"[list_index error]: {e}")
            return {
                "message": "Failed to list index",
                "error": str(e),
            }        



    def delete_index(self, file_id):
        """
        Delete all the index of a specific file_id, including its summary and chunks.
        """
        try:
            self.collection.delete(
                where={"file_id": file_id}
            )
            print(f"Deleted summary with file_id: {file_id}")
            return{
                "message": "Index deleted successfully",
                "file_id": file_id,
            }

        except Exception as e:
            print(f"[delete_index error]: {e}")
            return {
                "message": "Failed to delete index",
                "error": str(e),
            }


    def query_summaries_by_text(self, query_text, n_results=5) -> list:
        """
        Query summaries based on the query text.
        Args:
        - `query_text`: The text to query against.
        - `n_results`: Number of results to return.
        """
        try:  
            query_embedding = self.utils.get_embedding(query_text)

            raw_result = self.collection.query(
                query_embeddings=[query_embedding],
                n_results=n_results,
                where={"type": "summary"},
                include=["distances", "metadatas"]
            )

            ids = raw_result.get("ids", [[]])[0] if raw_result.get("ids") else []
            distances = raw_result.get("distances", [[]])[0] if raw_result.get("distances") else []
            metadatas = raw_result.get("metadatas", [[]])[0] if raw_result.get("metadatas") else []

            result_length = min(len(ids), len(distances), len(metadatas))

            flat_result = [
                {
                    "id": self._convert_numpy(ids[i]),
                    "distance": self._convert_numpy(distances[i]),
                    "metadata": self._convert_numpy(metadatas[i])
                }
                for i in range(result_length)
            ]

            return {
                "summaries": flat_result,
            }
        except Exception as e:
            print(f"[query_summaries_by_text error] {e}")
            return {
                "message": "Failed to query summaries",
                "error": str(e),
            }

    def query_chunks_by_text(self, query_text, file_id, n_results=5) -> list:
        """
        Query chunks that belong to a specific file_id, based on query text.
        Args:
        - `query_text`: The text to query against.
        - `file_id`: The ID of the file to filter chunks by.
        - `n_results`: Number of results to return.
        """
        try:
            query_embedding = self.utils.get_embedding(query_text)

            # Get the summary of the file_id
            summary_raw_result = self.collection.query(
                query_embeddings=[query_embedding],
                n_results=1,
                where={"$and": [{"type": "summary"}, {"file_id": file_id}]},
                include=["distances", "metadatas"]
            )
            
            summary_ids = summary_raw_result.get("ids", [[]])[0] if summary_raw_result.get("ids") else []
            summary_distances = summary_raw_result.get("distances", [[]])[0] if summary_raw_result.get("distances") else []
            summary_metadatas = summary_raw_result.get("metadatas", [[]])[0] if summary_raw_result.get("metadatas") else []

            summary_result_length = min(len(summary_ids), len(summary_distances), len(summary_metadatas))

            summary_flat_result = [
                {
                    "id": self._convert_numpy(summary_ids[i]),
                    "distance": self._convert_numpy(summary_distances[i]),
                    "metadata": self._convert_numpy(summary_metadatas[i])
                }
                for i in range(summary_result_length)
            ]

            print("The summary_flat_result is querying...")


            # Get the relevant chunks of the file_id
            raw_result = self.collection.query(
                query_embeddings=[query_embedding],
                n_results=n_results,
                where={"$and": [{"type": "chunk"}, {"file_id": file_id}]},
                include=["distances", "metadatas"]
            )

            ids = raw_result.get("ids", [[]])[0] if raw_result.get("ids") else []
            distances = raw_result.get("distances", [[]])[0] if raw_result.get("distances") else []
            metadatas = raw_result.get("metadatas", [[]])[0] if raw_result.get("metadatas") else []

            flat_result_length = min(len(ids), len(distances), len(metadatas))

            flat_result = [
                {
                    "id": self._convert_numpy(ids[i]),
                    "distance": self._convert_numpy(distances[i]),
                    "metadata": self._convert_numpy(metadatas[i])
                }
                for i in range(flat_result_length)
            ]

            print("The flat_result is querying...")

            # Return the summary and most relevant chunks of the file_id
            return {
                "file_id": file_id,
                "summary": summary_flat_result,
                "chunks": flat_result
            }
        except Exception as e:
            print(f"[query_chunks_by_text error] {e}")
            return {
                "message": "Failed to query chunks",
                "error": str(e),
            }
    


# __init__.py
from .hie_rag import HieRag
from .process import Process
from .split import Split
from .split_and_process import SplitAndProcess
from .tree_index import TreeIndex
from .utils import Utils
from .vectordb import Vectordb

__all__ = [
    "HieRag",
    "Process",
    "Split",
    "SplitAndProcess",
    "TreeIndex",
    "Utils",
    "Vectordb",
]


import os

from dotenv import load_dotenv
from hie_rag.app import Split
from hie_rag.process import Process
from hie_rag.utils import Utils
from hie_rag.vectordb import Vectordb

load_dotenv()

vectordb = Vectordb(path="test-vectordb", collection_name="index", api_key=os.getenv("OPENAI_API_KEY"))
split = Split(api_key=os.getenv("OPENAI_API_KEY"), min_chunk_size=200, max_chunk_size=500)
utils = Utils(api_key=os.getenv("OPENAI_API_KEY"))
process = Process(api_key=os.getenv("OPENAI_API_KEY"))

with open("test.pdf", "rb") as uploaded_file:
    extracted_text = utils.extract_text(uploaded_file)

# Prepare the chunks to process
extracted_text = extracted_text[:1000]
result_split = split.split(extracted_text)
result_process = process.process_chunks(result_split)


# Write results to the text file
with open("test-vectordb-result", "w", encoding="utf-8") as file:
    file.write("Save index:\n")
    file.write(str(vectordb.save_index(result_process)) + "\n")
    
    file.write("Query by text:\n")
    file.write(str(vectordb.query_by_text("臺中市政府")) + "\n")

    file.write("Query by embedding:\n")
    file.write(str(vectordb.query_by_embedding(utils.get_embedding("臺中市政府"))) + "\n")